#pragma once

#include "Node.h"
#include "Strategy.h"
#include <iostream>
class Tree
{
	Node* root;


	static void deleteSubtree(Node* curr) {
		if (curr->left) deleteSubtree(curr->left);
		if (curr->right) deleteSubtree(curr->right);
		delete curr;
	}

	static void printSubtree(std::ostream& os, Node* curr) {
		if (curr->left) printSubtree(os,curr->left);
		os << curr->val << " ";
		if (curr->right) printSubtree(os, curr->right);
	}

	static void traverseSubtree(const Strategy & strat, Node * curr) {
		if (curr->left) traverseSubtree(strat,curr->left);
		strat(curr);
		if (curr->right) traverseSubtree(strat,curr->right);
	}

	Tree() :root(nullptr) {}


	Tree& operator=(const Tree&) = delete;
	Tree(const Tree&) = delete;
public:


	~Tree() {
		deleteSubtree(root);
	}


	static Tree& getInstance() {
		static Tree t;
		return t;
	}



	void traverse(const Strategy & strat) {
		traverseSubtree(strat, root);
	}

	void traverseConst(const Strategy& strat) const {
		traverseSubtree(strat, root);
	}


	void insert(int val) {
		if (root == nullptr) {
			root = new Node(val);
			return;
		}
		Node* curr = root;
		while (curr) {
			if (val < curr->val) {
				if (curr->left)curr = curr->left;
				else {
					curr->left = new Node(val);
					return;
				}
			}
			else if (val > curr->val) {
				if (curr->right)curr = curr->right;
				else {
					curr->right = new Node(val);
					return;
				}
			}
			else {//equals
				return;
			}
		}
	}

	friend std::ostream& operator<<(std::ostream&, const Tree&);
};

